"""This modules provides the DiscreteExpression class to provide
common methods for the discrete-time representations.

Copyright 2020 Michael Hayes, UCECE

"""

from .expr import Expr


class DiscreteExpression(Expr):

    """Superclass of discrete-time, discrete-frequency, and z-domain
    expressions."""

    pass
